import tensorflow as tf
import numpy as np

from .custom_exceptions import MemThresholdReached

from .project_globals import TRAIN_TFR_NAME, EVAL_TFR_NAME, METADATA_FILE_NAME, TFR_IMPLICIT_DIR_NAME, TFR_SAMPLE_NAME, TFR_GROUND_THRUTH_NAME 
from .project_globals import DARKNET_DS_FLAG, TF_VERSION, CLASS_NAMES, IMAGE_W, IMAGE_H, COLOR, NUM_CLASSES, NUM_TRAIN_EXAMPLES, NUM_EVAL_EXAMPLES
from .project_globals import JSON_ANNOTATIO_DIR

from os import listdir, mkdir
from os.path import join, basename, dirname, isfile, isdir, splitext
from glob import glob
from time import sleep
from cv2 import imread, IMREAD_COLOR, IMREAD_GRAYSCALE, imshow, waitKey, resize
from itertools import zip_longest
from psutil import virtual_memory
from sys import stdout
from time import time
from json import load, dump


class ImgDataset():
        
    def __init__(self, ds_dir, img_extension, color, split_percent, memory_threshold):
    
        self._ds_dir           = ds_dir
        self._split_percent    = split_percent

        self._memory_threshold = memory_threshold

        self._img_extension    = img_extension
        self._color            = color
        
        self._imread_mode      = IMREAD_COLOR if color else IMREAD_GRAYSCALE
        
        ### populated after call to self._build()
        
        # maybe find a better whay than just poking at one sample image to get the _width and _height, which are the same for all images presumably
        self._width    = None
        self._height   = None
        self._channels = 3 if color else 1
        
        self._annot_ds_flag = None
        
        self._num_classes = 0
        self._class_names = [] 
        
        self._train_samples = []
        self._train_ground_truths = []
        self._eval_samples  = []
        self._eval_ground_thruths = []
    
        self._num_train_examples = 0
        self._num_eval_examples  = 0
        
        self._train_set_iterator = None
        self._eval_set_iterator  = None
        
        self._build()

    def is_annotation_dataset(self):
        
        return self._annot_ds_flag
    
    def get_ds_dir(self):
        
        return self._ds_dir
    
    def get_num_classes(self):
        
        return self._num_classes
    
    def get_num_train_examples(self):
        
        return self._num_train_examples
    
    def get_num_eval_examples(self):
        
        return self._num_eval_examples
    
    def get_width(self):
        
        return self._width
    
    def get_height(self):
        
        return self._height
    
    def get_color(self):
        
        return self._color
    
    def get_class_names(self):
        
        return self._class_names
    
    
    def get_dataset_metadata(self):
        
        return {
        
            
                    IMAGE_W    : self._width,
                    IMAGE_H    : self._height,
                    COLOR      : self._color,
                    NUM_CLASSES: self._num_classes,
            
                    NUM_TRAIN_EXAMPLES : self._num_train_examples,
                    NUM_EVAL_EXAMPLES  : self._num_eval_examples,
                    ANNOTATION_DS_FLAG : self._annot_ds_flag,
                    TF_VERSION         : tf.__version__,
                    CLASS_NAMES        : self._class_names
                        
                
            }
        
        
    def infer_dataset_type(self):
        
        all_samples = []
        all_labels = []
        all_txt_annot = []
        all_json_annot = []
        
        all_samples = glob(join(self.ds_dir, '*' + self._img_extension))
        
        if len(all_samples) == 0:
            
            self._annot_ds_flag = False
            
            # sample -- lists of lists of strings
            # labels -- lists of lists of integers
             
            all_samples, all_labels = self.load_dir(self.ds_dir)

        else:
            
            self._annot_ds_flag = True
            
            # list of strings
            
            all_txt_annot = glob(join(self.ds_dir, '*' + '.txt'))
            all_json_annot = None
            
            if len(all_txt_annot) == 0:
                
                # list of strings
                
                all_json_annot = glob(join(self.ds_dir, '*' + '.json'))
                all_txt_annot = None
                
                if len(all_json_annot) == 0:
                
                    self._annot_ds_flag = None # just in case exception is to be caught and info about dataset type is to be used, None being it's invalid
                    raise ValueError('Invalid dataset type.')
        
        return all_samples, all_labels, all_txt_annot, all_json_annot
        
    def find_sub_dir_paths(self, root_dir):
       
        class_dir_names = listdir( root_dir )
        class_dir_names.sort()

        self._class_names = class_dir_names
        self._num_classes = len( class_dir_names )

        return [ join( root_dir, class_dir_name ) for class_dir_name in class_dir_names ] 

    def find_file_paths(self, root_dir):
       
        return glob( join( root_dir, '*' + self._img_extension ) )

    def unison_shuffle(self, samples, labels):
        
        shuffled_samples = []
        shuffled_labels = []
        
        max_index = len(samples)
        
        while max_index > 0 :
        
            candidate_index = np.random.randint( 0, max_index )
            
            sample = samples.pop( candidate_index )
            label = labels.pop( candidate_index )
            
            shuffled_samples.append( sample )
            shuffled_labels.append( label )

            max_index = max_index - 1

        return shuffled_samples, shuffled_labels
    
    def split(self, all_class_samples, all_class_labels):
        
        tr_samples = []
        tr_labels  = []
        ev_samples = []
        ev_labels  = []
        
        for samples, labels in zip( all_class_samples, all_class_labels ) :
            
            total = len(samples)
            split_index = int( total * self._split_percent )
        
            tr_samples += samples[ 0: split_index ] 
            ev_samples += samples[ split_index : total - 1 ]  
        
            tr_labels += labels[ 0 : split_index ]  
            ev_labels += labels[ split_index: total - 1 ]  

        return tr_samples, tr_labels, ev_samples, ev_labels
    
    def load_dir(self, root_dir):
        
        class_paths = self.find_sub_dir_paths( root_dir )
        
        samples = []
        labels = []
        
        for i in range( self._num_classes ):
            
            path = class_paths[ i ]
            
            ith_class_samples = self.find_file_paths( path )
            
            nr_samples = len( ith_class_samples )
            ith_class_labels = [i] * nr_samples
            
            samples.append( ith_class_samples )
            labels.append( ith_class_labels )

        return samples, labels
    
    
    def parse_txt_annot_file(self, annot_file):
        
        image_annotations = []

        with open( annotations_file, 'r' ) as handler:
            
            lines = handler.readlines()
            
        content = [ line.strip() for line in lines ]
        
        for annotation in content:
            
            data = annotation.split()
            
            x = int( float( data[1] ) * self._width )
            y = int( float( data[2] ) * self._height)
            w = int( float( data[3] ) * self._width )
            h = int( float( data[4] ) * self._height)

            xmin = x - int( w / 2)
            xmax = x + int( w / 2)
            ymin = y - int( h / 2)
            ymax = y + int( h / 2)
            
            cls_index = int( data[0] )
        
            image_annotations.append( [ xmin, xmax, ymin, ymax, cls_index ] )
            
        return image_annotions
    
    def parse_json_annot_file(self, annot_file):
        
        pass
    
    
    def get_image(self, path):
    
        img = imread( path, self._imread_mode )
        
        if not self._color:
            
            img = np.expand_dims(img, 2)
    
        return img

    def get_ground_thruth(self, ground_thruth):

        '''
        
            parse darknet txt to get xmin xmax ymin ymax cls index
            
            or parse annot tool json to get - / -
            
            or get integer label
            
        '''
        
        if not self._annot_ds_flag:
            
            return ground_thruth
        
        else:
            
            if splitext(ground_thruth)[1] == '.txt':
                
                return self.parse_txt_annot_file(ground_thruth)
                
            else:
                
                return self.parse_json_annot_file(ground_thruth)

    
    def load_next_train_example(self):

        sample, ground_thruth = next( self._train_set_iterator )

        return self.get_image( sample ), self.get_ground_thruth( ground_thruth )
        
    def load_next_eval_example(self):
        
        sample, ground_thruth = next( self._eval_set_iterator )

        return self.get_image( sample ), self.get_ground_thruth( ground_thruth )

    def reset_train_set_iterator(self):
        
        self._train_set_iterator = zip_longest( self._train_samples, self._train_ground_truths )
        
    def reset_eval_set_iterator(self):
        
        self._eval_set_iterator  = zip_longest( self._eval_samples, self._eval_ground_thruths )


    def _build(self):

        all_samples, all_labels, all_txt_annot, all_json_annot = self.infer_dataset_type()
        
        if not self._annot_ds_flag:
        
            all_ground_thruths = all_labels

        else:
            
            '''
            
                Batch them up since self.split() is implemented to work with lists of lists, 
                I.E samples grouped by class, in order to get a uniform distribution of samples of each class
                in the train and evaluation set, by splitting each list from the list of lists by the given percentage.   
                
            '''
            
            all_samples = [all_samples] 
            
            if all_txt_annot is None:
                
                all_ground_thruths = [all_json_annot]
        
            else:
                
                all_ground_thruths = [all_txt_annot]
        
        tr_samples, tr_labels, ev_samples, ev_labels = self.split( all_samples, all_ground_thruths )
            
        self._train_samples, self._train_ground_truths = self.unison_shuffle( tr_samples, tr_labels )
        self._eval_samples, self._eval_ground_thruths   = self.unison_shuffle( ev_samples, ev_labels )
        
        self._train_set_iterator = zip_longest( self._train_samples, self._train_ground_truths )
        self._eval_set_iterator  = zip_longest( self._eval_samples, self._eval_ground_thruths )
        
        self._num_train_examples = len( self._train_samples )
        self._num_eval_examples  = len( self._eval_samples )
        
        # assuming dataset has been preped prior to using the main utility this should hold
        
        sample_image = imread(self._train_samples[0], self._imread_mode)
        self._width = sample_image.shape[1]
        self._height = sample_image.shape[0]
    

    def load_in_memory(self):
    
        train_images, eval_images = [], []
        train_ground_thruths, eval_ground_thruths = [], []
            
        # Loading train set ...
        try:
            
            while True:
                
                tr_image, tr_ground_thruth = self.load_next_train_example()
                train_images.append(tr_image)
                train_ground_thruths.append(tr_ground_thruth)
                
                if virtual_memory().percent > self._memory_threshold :
                
                    msg = 'Failed to load dataset into memory because it exceeds the user defined threshold of ' + str( self._memory_threshold ) + ' percent'
                    raise MemThresholdReached( msg )
            
        except StopIteration:
                
            print('End of train set examples.')
                
        except MemThresholdReached:
            
            self.reset_train_set_iterator() # converter will make use of iterator, might as well reset it now
        
        # Loading evaluation set ...
        try:
            
            while True:
                
                ev_image, ev_ground_thruth = self.load_next_eval_example()
                eval_images.append(ev_image)
                eval_ground_thruths.append(ev_ground_thruth)
                
                if virtual_memory().percent > self._memory_threshold :
                
                    msg = 'Failed to load dataset into memory because it exceeds the user defined threshold of ' + str( self._memory_threshold ) + ' percent.'
                    raise MemThresholdReached( msg )
            
        except StopIteration:
                
            print('End of eval set examples.')
        
        except MemThresholdReached:
            
            self.reset_eval_set_iterator() # converter will make use of iterator, might as well reset it now
            
        return ( 
            
            np.asarray( train_images, dtype = np.uint8 ), 
            np.asarray( train_ground_thruths, dtype = np.uint8), 
            np.asarray( eval_images, dtype = np.uint8), 
            np.asarray( eval_ground_thruths, dtype = np.uint8) )
        
class TFRGenerator():
    
    def __init__(self, img_dataset):
        
        self.img_dataset  = img_dataset
            
        self.tfr_root_dir = join( img_dataset.get_ds_dir(), TFR_IMPLICIT_DIR_NAME )
        
        
    def int64_feature(self, value):
        
        return tf.train.Feature( int64_list = tf.train.Int64List( value = [ value ] ) )
    
    def bytes_feature(self, value):
        
        return tf.train.Feature( bytes_list = tf.train.BytesList( value = [ value ] ) )
     
     
    def export_dataset_metadata(self, entry):
        
        metadata_file = join( self.tfr_root_dir, METADATA_FILE_NAME )
        
        if isfile( metadata_file ):
            
            with open( metadata_file, 'r' ) as json_file:
                
                json_data = load( json_file )
                json_data.update( entry )
                
            with open( metadata_file, 'w' ) as json_file:
                
                dump(json_data, json_file, indent = 10)    
            
        else: 
            
            json_data = entry
            
            with open(metadata_file, 'w') as json_file:
                
                dump(json_data, json_file, indent = 10)
    
    
    def build_feature(self, image, ground_thruth):
        
        if self.img_dataset.is_annotation_dataset():
            
            height, width, channels = image.shape
     
            nr_annotations = len(ground_thruth)
            annots = []
            
            for i in range(nr_annotations):
                
                annots += ground_thruth[ i ]
            
            annots += [nr_annotations]
            
            '''
                
                flattened version of the original [ [xmin, xmax, ymin, ymax, cls_index], [xmin, xmax, ymin, ymax, cls_index], ... ] with the last element 
                being the number of annotaitons, which is to be poped after deserialization in order to reshape the array to it's original shape.
            
            '''
            
            image_annotations = np.asarray( annots, dtype = np.uint8 )
            
            tfr_ground_thruth_feature = self.bytes_feature( tf.compat.as_bytes( image_annotations.tostring() ) )
        
        else:
            
            tfr_ground_thruth_feature = self.int64_feature(ground_truth)
        
        return {
            
            TFR_SAMPLE_NAME : self.bytes_feature( tf.compat.as_bytes( image.tostring() ) ), 
            TFR_GROUND_THRUTH_NAME  : tfr_ground_thruth_feature
            
            }
    
    
    def _convert(self, fetch_example_func, sub_set_name ):
        
        if not isdir( self.tfr_root_dir ): mkdir( self.tfr_root_dir )
        
        nr_examples = 0
        
        writer = tf.python_io.TFRecordWriter( join( self.tfr_root_dir, sub_set_name ) )
      
        try:
            
            while True:
            
                image, ground_thruth = fetch_example_func()
                nr_examples += 1

                feature = self.build_feature( image, ground_thruth )
                
                example = tf.train.Example( features = tf.train.Features( feature = feature ) )
            
                writer.write( example.SerializeToString() )
      
        except StopIteration:
            
            print('Finished serializing to ', sub_set_name)
            
        finally:
            
            writer.close()
            stdout.flush()
              
    def convert(self):

        start = time()
        
        self._convert( self.img_dataset.load_next_train_example, TRAIN_TFR_NAME )
        print ( 'Finished writing train tfr in: ', time() - start )
        
        start = time()
        
        self._convert( self.img_dataset.load_next_eval_example, EVAL_TFR_NAME )
        print ( 'Finished writing evaluation tfr in: ', time() - start )
    
    
        print ( 'TFR dataset found in directory: ' + self.tfr_root_dir, ' alongside the dataset metadata file.' )
        
        self.export_dataset_metadata( self.img_dataset.get_dataset_metadata() )
    
        return self.tfr_root_dir
        