import tensorflow as tf
import numpy as np
          
from .project_globals import TFR_SAMPLE_NAME, TFR_GROUND_THRUTH_NAME, BUTS_DIR_NAME, TFR_IMPLICIT_DIR_NAME
from .utils import bbox_iou

from os import listdir, mkdir
from os.path import dirname, join, basename, isfile, isdir
from random import uniform
from json import load
from glob import glob

from imgaug import Keypoint, KeypointsOnImage
from cv2 import imread, IMREAD_COLOR, IMREAD_GRAYSCALE, imwrite, resize, rectangle, putText, FONT_HERSHEY_SIMPLEX
from progressbar import ProgressBar, progressbar

class ImgPipe():
    
    # THIS NEEDS TO KNOW WETHER IT's a CLASSIFIER OR NOT, in order to decided wether to crop annotation ground thruths
    # and and make the actual images that are to be used in the training loop, after augs are finished
    
    def __init__(self, aug_pipe, params, num_classes ):
        
        self.aug_pipe     = aug_pipe
        self.params       = params
        self.num_classes  = num_classes
        
        self.label_dtype  = tf.uint8
        self.sample_dtype = tf.uint8
        
        self.features = { 
                 
            TFR_SAMPLE_NAME : tf.FixedLenFeature( [], tf.string ), 
            TFR_GROUND_THRUTH_NAME  : tf.FixedLenFeature( [], tf.int64 )  
                    
            }


    def _build_sample(self, raw_sample, sample_shape):
        
        sample = tf.decode_raw( raw_sample, self.sample_dtype )
        
        return tf.reshape( sample, sample_shape )
    
    def _build_label(self, raw_label):
        
        label = tf.cast( raw_label, tf.int64 )
        
        return tf.one_hot( 
            
                    indices   = label, 
                    depth     = self.num_classes, 
                    on_value  = tf.cast(1, self.label_dtype), 
                    off_value = tf.cast(0, self.label_dtype), 
                    dtype     = self.label_dtype
                    
                )
    
    def _deserialize(self, serialized, sample_shape):

        parsed_example = tf.parse_single_example( serialized, self.features )
        
        sample = self._build_sample( parsed_example[ TFR_SAMPLE_NAME ], sample_shape )
        label  = self._build_label( parsed_example[ TFR_LABEL_NAME ] )
        
        return sample, label
    

    
    def _augment_batch_wrapper(self):
        
        # tf py_func cant convert py objects like aug_pipe to tensor, that's why we need this wrapper so that the object is in scope and ready to be used in the actual function, which is defined inside the wrapper func
        
        def _augment_batch(images, labels):
                
            det_aug = self.aug_pipe.to_deterministic()
            
            return det_aug.augment_image( images ), labels
    
        return _augment_batch
    
    def _build_aug_py_func(self, img, lbl):

        py_func_op = tf.py_func( func = self._augment_batch_wrapper(), inp = [ img, lbl ], Tout = ( np.uint8, np.uint8 ) )
        
        py_func_op[0].set_shape( img.get_shape() )
        
        return py_func_op
    

    
    def _build_pipe(self, dataset, train_flag):
        
        if train_flag:
            
            dataset = dataset.map( lambda img, lbl: self._build_aug_py_func( img, lbl ), self.params['NR_CPU_CORES'] )
            dataset = dataset.shuffle( buffer_size = self.params['SHUFFLE_BUFFER_SIZE'] )
            dataset = dataset.prefetch( self.params['PREFETCH_POOL_SIZE'] )
            num_repeat = None
        
        else :
            
            num_repeat = 1
    
        dataset = dataset.map( lambda img, lbl: ( tf.divide( tf.cast(img, tf.float32), 255. ), lbl ), self.params['NR_CPU_CORES'] )
        
        dataset = dataset.repeat( num_repeat )
        dataset = dataset.batch( batch_size = self.params['BATCH_SIZE'] )
        dataset = dataset.prefetch( self.params['PREFETCH_POOL_SIZE'] )
        
        iterator = dataset.make_one_shot_iterator()
         
        return iterator.get_next()



    def build_tfr_pipe(self, path_to_tfr_file, sample_shape, train_flag):
        
        '''
    
            invoke when catching the MemThreshold exception
    
        '''
        
        dataset = tf.data.TFRecordDataset( path_to_tfr_file )
 
        dataset = dataset.map( lambda serialized: self._deserialize( serialized, sample_shape ), self.params['NR_CPU_CORES'] )
              
        dataset = dataset.prefetch( self.params['PREFETCH_POOL_SIZE'] ) 
        
        return self._build_pipe( dataset, train_flag )


    def build_mem_pipe(self, samples, labels, train_flag):
        
        '''
        
            invoke after loading dataset in memory procedure was succesfull 
        
        '''
        
        
        return self._build_pipe( tf.data.Dataset.from_tensor_slices( ( samples, labels ) ), train_flag )
        
class YoloPipe(ImgPipe):
    
    def __init__(self, aug_pipe, params, num_classes, yolo_spec):
        
        super().__init__( aug_pipe, params, num_classes )
        
        self.features = { 
            
            TFR_SAMPLE_NAME : tf.FixedLenFeature( [], tf.string ), 
            TFR_LABEL_NAME  : tf.FixedLenFeature( [], tf.string )
                     
            }

        self.yolo_spec = yolo_spec
        self.elements_per_annotation = 5 


    def _build_label(self, raw_label):
        
        '''
        
            annotations for one image --> [ [xmin, xmax, ymin, ymax, cls_index], [xmin, xmax, ymin, ymax, cls_index], [xmin, xmax, ymin, ymax, cls_index], ... ]
            
            * after decode we get the flattened array, of which the last element is the number of annotations
            * poping last element and using it to reshape the remaining array to [num_annotations, elements_per_annotation]
        
        '''
        
        annotations     = tf.decode_raw( raw_label, self.label_dtype )
        num_elements    = tf.size( annotations )
        num_annotations = annotations[ num_elements - 1 ]
        annotations     = tf.slice( annotations, [0], [ num_elements - 1 ] )
        
        # apperantly only int32 is accepted as integer args in the shape tuple, might have to look into this to see why
        return tf.reshape( annotations, [ tf.cast( num_annotations, tf.int32 ), tf.cast( self.elements_per_annotation, tf.int32 ) ] )


    def _augment_batch_wrapper(self):
        
        def _augment_batch(images, labels):
                
            det_aug = self.aug_pipe.to_deterministic()
            
            images = det_aug.augment_image( images )
            
            image_shape = images.shape
            
            im_width  = image_shape[1]
            im_height = image_shape[0]
            
            keypoints_on_images = []
            keypoints = []
             
            # for each image, labels = [ ann1 -> [xmin, xmax, ymin, ymax, cls_index], ann2 -> [xmin, xmax, ymin, ymax, cls_index], ... ]
             
            for label in labels:
                 
                keypoints.append( Keypoint( x = label[0], y = label[2] ) ) # top left
                keypoints.append( Keypoint( x = label[1], y = label[3] ) ) # bottom right
                keypoints.append( Keypoint( x = label[0], y = label[3] ) ) # bottom left
                keypoints.append( Keypoint( x = label[1], y = label[2] ) ) # top right
        
            keypoints_on_images.append( KeypointsOnImage( keypoints, shape = image_shape ) )
            keypoints_on_images = det_aug.augment_keypoints( keypoints_on_images )
        
            index = 0
            for keypoint in keypoints_on_images[0].keypoints:
                 
                if index % 4 == 0:
                    
                    x1, y1 = keypoint.x, keypoint.y
                     
                if index % 4 == 1:
                    
                    x2, y2 = keypoint.x, keypoint.y
         
                if index % 4 == 2:
                    
                    x3, y3 = keypoint.x, keypoint.y
                    
                if index % 4 == 3:
                    
                    x4, y4 = keypoint.x, keypoint.y
                    
                    xmin = min( x1, x2, x3, x4 )
                    xmax = max( x1, x2, x3, x4 )
                    ymin = min( y1, y2, y3, y4 )
                    ymax = max( y1, y2, y3, y4 )
        
                    labels[ int( index / 4 ) ][0] = xmin
                    labels[ int( index / 4 ) ][1] = xmax
                    labels[ int( index / 4 ) ][2] = ymin
                    labels[ int( index / 4 ) ][3] = ymax
                     
                index += 1
    
            return images, labels
    
        return _augment_batch


    def _yolo_batching_wrapper(self):

        def _yolo_batching(image, label):
        
            im_width = image.shape[1]
            im_height = image.shape[0]
        
            true_box_index = 0
            
            b_label = np.zeros( ( 1, 1, 1, self.yolo_spec['TRUE_BOX_BUFFER'], 4 ), dtype = np.float32 )    # list of self.config['TRUE_self.config['BOX']_BUFFER'] GT boxes
            y_label = np.zeros( ( self.yolo_spec['GRID_H'], self.yolo_spec['GRID_W'], int( len( self.yolo_spec['ANCHORS'] ) / 2 ), 4 + 1 + self.num_classes ), dtype = np.float32 ) # desired network output
            
            # xmin, ymin, xmax, ymax is the original implementation using BoundBox object 
            anchors = [ [0, 0, self.yolo_spec['ANCHORS'][2 * i], self.yolo_spec['ANCHORS'][ 2 * i + 1 ] ] for i in range( int( len( self.yolo_spec['ANCHORS'] ) // 2 ) ) ]
            
            label = label.astype(np.int32) # if label dtype == uint8 --> overflow on arithmetic ops like ( xmin + xmax ) * 0.5 wtfyayo
            
            for label in label:
                
                xmin      = label[0]
                xmax      = label[1]
                ymin      = label[2]
                ymax      = label[3]
                cls_index = label[4]
                
                # dunno why this condition is needed, don't see how this will fail, maybe double checking for conversion from relative to absolute bbox coords !?
                
                if xmax > xmin and ymax > ymin:
                    
                    center_x = .5 * ( xmin + xmax )
                    center_x = center_x / ( float( im_width ) / self.yolo_spec['GRID_W'] )
                    
                    center_y = .5 * ( ymin + ymax )
                    center_y = center_y / ( float( im_height ) / self.yolo_spec['GRID_H'] )
            
                    grid_x = int( np.floor( center_x ) )
                    grid_y = int( np.floor( center_y ) )
            
                    if grid_x < self.yolo_spec['GRID_W'] and grid_y < self.yolo_spec['GRID_H']:
                        
                        center_w = ( xmax - xmin ) / ( float( im_width ) / self.yolo_spec['GRID_W'] ) # unit: grid cell
                        center_h = ( ymax - ymin ) / ( float( im_height ) / self.yolo_spec['GRID_H'] ) # unit: grid cell
                        
                        bbox = [ center_x, center_y, center_w, center_h ]
            
                        # find the anchor that best predicts this bbox
                        best_anchor = -1
                        max_iou     = -1
            
                        shifted_box = [0, 0, center_w, center_h ]
                        
                        for i in range( len( anchors ) ):
                            
                            anchor = anchors[i]
                            iou    = bbox_iou( shifted_box, anchor )
                            
                            if max_iou < iou:
                                
                                best_anchor = i
                                max_iou     = iou
                                
                        # assign ground truth x, y, w, h, confidence and class probs to y_label
                        y_label[ grid_y, grid_x, best_anchor, 0:4 ] = bbox
                        y_label[ grid_y, grid_x, best_anchor, 4  ] = 1.
                        y_label[ grid_y, grid_x, best_anchor, 5 + cls_index ] = 1
                        
                        # assign the true bbox to b_label
                        b_label[0, 0, 0, true_box_index] = bbox
                        
                        true_box_index += 1
                        true_box_index = true_box_index % self.yolo_spec['TRUE_BOX_BUFFER']
                        
            return image.astype(np.float32), y_label, b_label
    
        return _yolo_batching

    def _build_yolo_batching_py_func(self, image, label ):
    
        py_func_op = tf.py_func( func = self._yolo_batching_wrapper(), inp = [ image, label ], Tout = ( np.float32, np.float32, np.float32 ) )
        py_func_op[0].set_shape( image.shape )
        
        return py_func_op



    def _build_pipe(self, dataset, train_flag):
        
        if train_flag:
            
            dataset = dataset.map( lambda img, lbl: self._build_aug_py_func( img, lbl ), self.params['NR_CPU_CORES'] )
            dataset = dataset.shuffle( buffer_size = self.params['SHUFFLE_BUFFER_SIZE'] )
            dataset = dataset.prefetch( self.params['PREFETCH_POOL_SIZE'] )
            num_repeat = None
        
        else :
            
            num_repeat = 1
    
        dataset = dataset.map( lambda img, lbl: self._build_yolo_batching_py_func( img, lbl ), self.params['NR_CPU_CORES'] )
        dataset = dataset.map( lambda img, y_lbl, b_lbl: ( tf.divide( img, 255. ), y_lbl, b_lbl ), self.params['NR_CPU_CORES'] )
    
        dataset = dataset.repeat( num_repeat )
        dataset = dataset.batch( batch_size = self.params['BATCH_SIZE'] )
        dataset = dataset.prefetch( self.params['PREFETCH_POOL_SIZE'] )
        
        iterator = dataset.make_one_shot_iterator()
        
        img, y_lbl, b_lbl = iterator.get_next()
        
        return img, (y_lbl, b_lbl)



class ClassifierEstimator(tf.estimator.Estimator):

    
    def _dir_check(self, dir):
        
        # maybe bundle up the dir path construction also, using os.path.join, removing some of the duplicate code in predict_wrapper
        if not isdir(dir): mkdir(dir)
    
    def _get_generatorX(self, image_files, color):
        
        imread_mode = IMREAD_COLOR if color else IMREAD_GRAYSCALE
        
        def _build_generatorX():
        
            for image_file in image_files:
                
                normalized_image = imread( image_file, imread_mode )
                normalized_image = normalized_image.astype( np.float32 ) / 255.
                
                if not color:
                    
                    normalized_image = np.expand_dims( normalized_image, 2 ) # account for 2 dim GRAYSCALE read, (width, height) only
                
                yield np.expand_dims(normalized_image, 0) # batch it up, the input for the estimator model must be 4 dim, I.E. batch dim included
                
        return _build_generatorX
                
    def _build_predict_input_fnX(self, image_files, width, height, color):

        # return a callable for the estimator api to use
        return lambda : tf.data.Dataset.from_generator( self._get_generatorX( image_files, color ), output_types = tf.float32, output_shapes = tf.TensorShape( [None, width, height, 3 if color else 1] ) )
     
     
    def predict_wrapper(self, src_dir, dst_dir, img_extension, train_info):
    
        image_files  = [] # string list, paths to image files
        labels       = [] # integer list, class index in cls prob distribution 
        
        imread_mode = IMREAD_COLOR if train_info['COLOR'] else IMREAD_GRAYSCALE
        
        class_label_dirs = listdir( src_dir )
        
        num_class_labels = len( class_label_dirs )
        
        num_removed = 0
        
        # remove potential hidden files and/or folders, or the implicit TFR_DATASET folder that might be present due to not being able to load dataset into memory
        for i in range( num_class_labels ):
        
            actual_index = i - num_removed
            class_label_dir = class_label_dirs[ actual_index ]
        
            if class_label_dir.startswith('.') or ( class_label_dir == TFR_IMPLICIT_DIR_NAME ):
        
                num_removed += 1
                class_label_dirs.pop( actual_index )

        if not class_label_dirs:
            
            image_files = glob( join( src_dir, '*' + img_extension ) )
            
        else:
            
            class_label_dirs.sort()
            
            num_classes = len( class_label_dirs )
            
            for i in range( num_classes ):
                
                ith_class_image_files = glob( join( join( src_dir, class_label_dirs[i] ), '*' + img_extension ) )
                
                nr_samples = len( ith_class_image_files )
                ith_class_labels = [i] * nr_samples
    
                image_files += ith_class_image_files
                labels += ith_class_labels
    
    
        prediction_generator = self.predict( self._build_predict_input_fnX( image_files, train_info['WIDTH'], train_info['HEIGHT'], train_info['COLOR'] ), yield_single_examples = True  )
        
        net_outputs = []
        num_examples = 0
        
        try:
            
            print('Computing predictions for all images in source directory ... ')
            
            while True:
                
                    net_outputs.append( next(prediction_generator) )
                    num_examples += 1
                
        except StopIteration:
                
            print ('Finished computing predictions for ', num_examples, ' images')
            
        print('Writing images to destination directory based on the predicted class ...')
            
        predictions = []

        for i in progressbar( range( num_examples ) ):
            
            image_file = image_files[i]
            net_out    = net_outputs[i]
            
            predicted_class_index = net_out.argmax(axis = 0)
            
            image = imread(image_file, imread_mode)
            
            class_label     = train_info['CLASS_NAMES'][predicted_class_index]
            image_file_name = basename(image_file)

            if labels:
                
                match = predicted_class_index == labels[i]
                predictions.append( match )
                
                if match:
                    
                    pred_dst_dir = join(dst_dir, class_label)
                    
                    self._dir_check(pred_dst_dir)
                    
                    pred_dst_file = join( pred_dst_dir, image_file_name )

                    if not imwrite(pred_dst_file, image):
                        
                        print('Failed to write image to ', pred_dst_file)
                        
                else:
                    
                    
                    buts_dir = join(dst_dir, BUTS_DIR_NAME)
                    self._dir_check(buts_dir)
                    
                    but_dir_name = train_info['CLASS_NAMES'][predicted_class_index] + ' BUT ' + train_info['CLASS_NAMES'][ labels[i] ]
                    
                    specific_but_dir = join(buts_dir, but_dir_name)
                    self._dir_check(specific_but_dir)
                
                    specific_but_file = join(specific_but_dir, image_file_name )
                
                    if not imwrite(specific_but_file, image):
                        
                        print('Failed to write image to ', specific_but_file)
                        
            else:
                
                pred_dst_dir = join(dst_dir, class_label)
                self._dir_check(pred_dst_dir)
                
                pred_dst_file = join( pred_dst_dir, image_file_name )
                
                if not imwrite(pred_dst_file, image):
                    
                    print('Failed to write image to ', pred_dst_file)

        predictions = np.asarray(predictions, dtype = np.float32)
        
        accuracy = np.mean(predictions, axis = 0)
                
        with open( join(dst_dir, 'ACCURACY: ' + str(accuracy) ), 'w' ) as dummy_handler:
                    
            # just wanted to create an empty file that will have a name that will encode the accuracy metric for, easy lookup, in dst_dir
            pass


class YoloEstimator(ClassifierEstimator):
    
    def _sigmoid(self, x):
        
        return 1. / (1. + np.exp(-x))

    def _softmax(self, x, axis = -1, t = -100.):
        
        x = x - np.max(x)
        
        if np.min(x) < t:
            
            x = x/np.min(x)*t
            
        e_x = np.exp(x)
        
        return e_x / e_x.sum(axis, keepdims = True)

    def _draw_boxes(self, image, bboxes, color = (0, 255, 0), y_shift = 10):
        
        image_h, image_w, _ = image.shape

        for bbox in bboxes :
            
            xmin = int( bbox[0] * image_w )
            ymin = int( bbox[1] * image_h )
            xmax = int( bbox[2] * image_w )
            ymax = int( bbox[3] * image_h )
    
            rectangle(image, (xmin, ymin), (xmax, ymax), color, 2)
            
            putText(image, 
                        str( np.argmax( bbox[..., 5:] ) ) + ' ' + str( bbox[4] ), 
                        ( xmin, ymin - y_shift ), 
                        FONT_HERSHEY_SIMPLEX, 
                        1e-3 * image_h, 
                        color, 2)
            
        return image
        

    def _decode_netout(self, net_out, num_classes, yolo_spec):
        
        grid_h, grid_w, nb_box = net_out.shape[: 3 ]

        boxes = []
        
        net_out[ ..., 4 ]  = self._sigmoid( net_out[..., 4] )
        net_out[ ..., 5: ] = self._softmax( net_out[..., 5:] )
        
        for row in range( grid_h ):
            
            for col in range( grid_w ):
                
                for b in range( nb_box ):
                    
                    classes = net_out[ row, col, b, 5: ]
                    
                    x, y, w, h = net_out[ row, col, b, :4]
    
                    x = ( col + self._sigmoid( x ) ) / grid_w 
                    y = ( row + self._sigmoid( y ) ) / grid_h
                    
                    w = yolo_spec['ANCHORS'][ 2 * b + 0 ] * np.exp( w ) / grid_w
                    h = yolo_spec['ANCHORS'][ 2 * b + 1 ] * np.exp( h ) / grid_h
                    
                    confidence = net_out[ row, col, b, 4 ]
    
                    if confidence > yolo_spec['OBJ_THRESHOLD']:
                        
                        # bbox -> [ xmin, ymin, xmax, ymax, confidence, probability]
                            
                        bbox = np.append( [ x - w / 2, y - h / 2, x + w / 2, y + h / 2, confidence ], classes ).astype( np.float32 )              
                        boxes.append( bbox )
    
        for c in range( num_classes ):
            
            sorted_indices = list( reversed( np.argsort( [ bbox[ ..., 5: ][ c ] for bbox in boxes ] ) ) ) 
    
            num_sorted_indices = len( sorted_indices )
    
            for i in range( num_sorted_indices ):
                
                index_i = sorted_indices[ i ]
                
                if boxes[ index_i ][..., 5:][c] == 0 or boxes[ index_i ][ 4 ] <= yolo_spec['OBJ_THRESHOLD']:
                     
                    continue
                
                else:
                    
                    for j in range( i + 1, num_sorted_indices ):
                        
                        index_j = sorted_indices[ j ]
                        
                        if  bbox_iou( boxes[ index_i ], boxes[ index_j ] ) >= yolo_spec['NMS_THRESHOLD']:
                            
                            if np.argmax( boxes[ index_j ][..., 5:] ) == c:  
                                                                              
                                boxes[ index_j ][ 4 ] = 0
                                
                            boxes[index_j][..., 5:][ c ] = 0
                            
                            
        boxes = [ bbox for bbox in boxes if bbox[ 4 ] > yolo_spec['OBJ_THRESHOLD'] ]
        
        return boxes
        

    def predict_wrapper(self, src_dir, dst_dir, img_extension, train_info, yolo_spec):
        
        image_files = glob( join( src_dir, '*' + img_extension ) )
        
        prediction_generator = self.predict( 
                                            
                                        self._build_predict_input_fnX( 
                                                                           
                                                            image_files, 
                                                            train_info['WIDTH'], 
                                                            train_info['HEIGHT'], 
                                                            train_info['COLOR'] ),
                                             
                                        yield_single_examples = True  
                                        
                                    )
        
        net_outputs = []
        num_examples = 0
        
        try:
            
            while True:
        
                net_outputs.append( next(prediction_generator) )
                num_examples += 1
        
        except StopIteration:
            
            print ('Finished computing predictions for ', num_examples, ' images')
        
        progress_bar = ProgressBar()
        
        for image_file, net_out in zip(progress_bar(image_files), net_outputs):
            
            image = imread( image_file )
            
            bboxes = self._decode_netout( net_out, train_info['NUM_CLASSES'], yolo_spec )
            
            image = resize( image, (400, 400) ) 
            
            annotated_image = self._draw_boxes(image, bboxes)
            
            ground_truth = []
            
            annotation_file = image_file[:-4] + '.txt'
            
            if isfile( annotation_file ):

                with open( annotation_file, 'r' ) as handler:
                    
                    lines = handler.readlines()
                    
                content = [ line.strip() for line in lines ]
                
                for annotation in content:
                    
                    data = annotation.split()
                    
                    x = float( data[1] )
                    y = float( data[2] )
                    w = float( data[3] )
                    h = float( data[4] ) 
        
                    # still relative coordinates, will get absolute coords depending on what the final image on which bboxes will be drawn will be, same thing comes out of the network
                    
                    xmin = x - w / 2
                    xmax = x + w / 2
                    ymin = y - h / 2
                    ymax = y + h / 2 
        
                    cls_index = int( data[0] )
                
                    class_prob_distribution = np.zeros( [ train_info['NUM_CLASSES'] ], dtype = np.float32 )
                    class_prob_distribution[ cls_index ] = 1
                    
                    ground_truth.append( np.append( np.asarray( [ xmin, ymin, xmax, ymax, 1 ], dtype = np.float32 ), class_prob_distribution ) ) 
                    
                # TODO: find a better way to depict class label as text on the image, which atm is just class index in prob dist net_out, and object prob, in some cases might end up being outside the image, I.E top side of the image
                annotated_image = self._draw_boxes( image, ground_truth, color = (0, 0, 255), y_shift = 20 )
            
            # TODO: decide what will be the sub dir inside dst_dir in which the annotaed_image will go, for instance, FALSE_POSITIVE, FALSE_NEGATIVE or in the root dir itself if TRUE POSITIVES
            
            imwrite( join( dst_dir, basename( image_file ) ), annotated_image )
            
        

