# &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& 

import tensorflow as tf
import logging

from argparse import ArgumentParser
from os.path import join, basename, dirname
from json import load, dump
from platform import platform
from sys import path as PYTHONPATH

from .dataset_utils import ImgDataset, YoloDataset, ImgConverter, YoloConverter

from .estimator_prep import ImgPipe, YoloPipe, ClassifierEstimator, YoloEstimator

from .utils import get_max_pool_count, build_estimator_model_function, update_config, check_args_validity, export_inference_model

from .custom_exceptions import MemThresholdReached

from .project_globals import TRAIN, PREDICT, EXPORT, CLS, DTT
from .project_globals import TRAIN_TFR_NAME, EVAL_TFR_NAME, CONFIG_FILE_NAME, MODEL_FILE_NAME, METADATA_FILE_NAME
from .project_globals import MODE_ARG, TASK_ARG, CONFIG_ARG, DUMP_ARG, PLATTEAU_TOLERANCE_ARG, IMG_DATASET_ARG, COLOR_ARG, EXTENSION_ARG
from .project_globals import PERCENT_ARG, THRESHOLD_ARG, TFR_DATASET_ARG, CHECKPOINT_DIR_ARG, EXPORT_MODEL_DIR_ARG, EXPORT_MODEL_FILE_NAME_ARG
from .project_globals import SRC_DIR_ARG, DST_DIR_ARG
from .project_globals import TRAIN_DS_PERCENT_MIN, TRAIN_DS_PERCENT_MAX, MEM_THRESHOLD_PERCENT_MIN, MEM_THRESHOLD_PERCENT_MAX
from .project_globals import ALLOWED_EXTENSIONS

# &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&

parser = ArgumentParser()

parser.add_argument(MODE_ARG, choices = [TRAIN, PREDICT, EXPORT], help = 'mode of operation for the utility')

### TRAIN and PREDICT args

parser.add_argument('--' + TASK_ARG, choices = [CLS, DTT], help = 'wether to work with a classifier or yolo detector')
parser.add_argument('--' + CONFIG_ARG, help = 'path to configuration root directory, where config.json and model.py is to be found')
parser.add_argument('--' + DUMP_ARG, help = 'path to model checkpoints and tensorboard logging dump directory')

parser.add_argument(
                
                '--' + PLATTEAU_TOLERANCE_ARG, 
                type = int, 
                help = 'DO NOT USE THIS PARAM AT THE MOMENT, DANGER DANGER !!!'
                    
            )


'''

    Note:  
    
        Might be a bit redundant to have two args for dataset paths, 
        better to infer wether its img or tfr from the presence of the required args for an img dataset. MAYBE !!?


'''

parser.add_argument('--' + IMG_DATASET_ARG, help = 'Path to image dataset root dir ')
parser.add_argument('--' + COLOR_ARG, type = int, choices = [1, 0], help = 'Flag for when to read images, COLOR_BGR or GRAYSCALE')
parser.add_argument('--' + EXTENSION_ARG, choices = ALLOWED_EXTENSIONS, help = 'Image extension to look out for')
parser.add_argument(
                    
                '--' + PERCENT_ARG, 
                type = float, 
                help = 'Train dataset percentage, float between ' + str( TRAIN_DS_PERCENT_MIN ) + ' and ' + str( TRAIN_DS_PERCENT_MAX )
                
            )
                
parser.add_argument(
                    
                '--' + THRESHOLD_ARG, 
                type = float, 
                help = 'Memory threshold cap for loading dataset, float between ' + str(MEM_THRESHOLD_PERCENT_MIN) + ' and ' + str(MEM_THRESHOLD_PERCENT_MAX) )

parser.add_argument('--' + TFR_DATASET_ARG, help = 'Path to tfr dataset root dir')

parser.add_argument('--' + SRC_DIR_ARG, help = 'Source directory for images that are to be used for ' + PREDICT)
parser.add_argument('--' + DST_DIR_ARG, help = 'Destination directory for ' + PREDICT)

###

### EXPORT args

parser.add_argument(
                    
    '--' + CHECKPOINT_DIR_ARG, 
    help = 'Source directory where checkpoints are to be found, at the moment the most recent checkpoint will be taken, since in training with this utility, chackpoints are saved on a save best only basis'
                
)

parser.add_argument('--' + EXPORT_MODEL_DIR_ARG, help = 'Destination directory where the inference model will be dumped')
parser.add_argument('--' + EXPORT_MODEL_FILE_NAME_ARG, help = 'String ending in .pb')

###

arguments = vars ( parser.parse_args() )

checked_arguments = check_args_validity(arguments)

( MODE, TASK, CONFIG_DIR, DUMP_DIR, PLATTEAU_TOLERANCE, IMG_DATASET, 
  COLOR, EXTENSION, PERCENT, THRESHOLD, TFR_DATASET, SRC_DIR, DST_DIR, 
  CHECKPOINT_DIR, EXPORT_MODEL_DIR, EXPORT_MODEL_FILE_NAME ) = checked_arguments 
  
# &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& 

if MODE == EXPORT:
    
    export_inference_model(CHECKPOINT_DIR, EXPORT_MODEL_DIR, EXPORT_MODEL_FILE_NAME) # script stops here if EXPORT is the intention

CONFIG_FILE = join( CONFIG_DIR, CONFIG_FILE_NAME )

if MODE == TRAIN:

    if IMG_DATASET is not None:
        
        if TASK == DTT: image_dataset = YoloDataset( IMG_DATASET, EXTENSION, COLOR, PERCENT, THRESHOLD )
            
        else: image_dataset = ImgDataset( IMG_DATASET, EXTENSION, COLOR, PERCENT, THRESHOLD )
    
        NUM_TRAIN_EXAMPLES = image_dataset.get_num_train_examples() 
        NUM_EVAL_EXAMPLES  = image_dataset.get_num_eval_examples()
        NUM_CLASSES        = image_dataset.get_num_classes()
        CLASS_NAMES        = image_dataset.get_class_names()
    
        WIDTH  = image_dataset.get_width()
        HEIGHT = image_dataset.get_height()
        COLOR  = image_dataset.get_color()
            
    else : # I.E. TFR dataset
        
        with open( join( TFR_DATASET, METADATA_FILE_NAME ), 'r') as handler:
            
            dataset_metadata = load(handler)
        
        NUM_TRAIN_EXAMPLES = dataset_metadata['NUM_TRAIN_EXAMPLES']
        NUM_EVAL_EXAMPLES  = dataset_metadata['NUM_EVAL_EXAMPLES']
        NUM_CLASSES        = dataset_metadata['NUM_CLASSES']
        CLASS_NAMES        = dataset_metadata['CLASS_NAMES']
    
        WIDTH  = dataset_metadata['IMAGE_W']
        HEIGHT = dataset_metadata['IMAGE_H']
        COLOR  = dataset_metadata['COLOR']

else : # I.E PREDICT since EXPORT doesn't even reach this far
    
    
    with open(CONFIG_FILE, 'r') as handler:
        
        CONFIG = load(handler)
    
    try:
        
        WIDTH  = CONFIG['TRAIN_INFO']['WIDTH']
        HEIGHT = CONFIG['TRAIN_INFO']['HEIGHT']
        COLOR  = CONFIG['TRAIN_INFO']['COLOR']
        
    except KeyError:
        
        '''
            
            Note:
            
            Maybe do some checking when it comes to tensorflow version used in training also, that might have an effect on how the predictions go
            Same thing for EXPORT mode
        
        '''
        
        
        print('The config.json file provided was not used to train a model and filling in the blanks with what you might think the training specifics were might lead to different results in predictions')

    YOLO_SPEC     = CONFIG['YOLO_SPEC']

    TRAIN_CONFIG  = CONFIG['TRAIN_CONFIG']

    BATCHING_SPEC = TRAIN_CONFIG['BATCHING']
    LOGGING       = TRAIN_CONFIG['LOGGING']
    
    NUM_CLASSES = CONFIG['TRAIN_INFO']['NUM_CLASSES']
    TRAIN_STEPS = None
    

# &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&& 

PYTHONPATH.insert( 0, CONFIG_DIR )

from model import *

INPUT_SHAPE = [ WIDTH, HEIGHT, 3 if COLOR else 1 ]

user_defined_model_function  = locals()['model_function']
user_defined_aug_pipe        = locals()['AUG_PIPE']


if MODE == TRAIN:

    if TASK == DTT : # load user defined graph, count max pool ops, compute GRID_W and GRID_H, update initial config with this info 

        dummy_input = tf.placeholder('float', [ None, WIDTH, HEIGHT, 3 if COLOR else 1 ], name = 'dummy_input')

        # No need for the returned output_op and optimizer_op just yet, doing this solely for counting max pool ops and updating config
        # Even the dummy_input and bool args passed do not have any effect whatsoever on the procsess
        
        _, _ = user_defined_model_function( dummy_input, True )
        num_max_pools = get_max_pool_count( tf.get_default_graph() )
    
        GRID_W = int( WIDTH / 2 ** num_max_pools )
        GRID_H = int( HEIGHT / 2 ** num_max_pools )
        
        _ = update_config( CONFIG_FILE, { 'GRID_W': GRID_W, 'GRID_H': GRID_H }, key = 'YOLO_SPEC' )
        
    
    # updating config file with training specifics either derived from the bound dataset or technology specifics like tf version which is also relevant
    NEW_CONFIG = update_config( CONFIG_FILE, { 'TRAIN_INFO': { 
                        
                                            'TASK'                    : TASK,
                                            'WIDTH'                   : WIDTH, 
                                            'HEIGHT'                  : HEIGHT, 
                                            'COLOR'                   : COLOR, 
                                            'NUM_TRAIN_EXAMPLES'      : NUM_TRAIN_EXAMPLES, 
                                            'NUM_EVAL_EXAMPLES'       : NUM_EVAL_EXAMPLES, 
                                            'NUM_CLASSES'             : NUM_CLASSES,
                                            'TF_VERSION'              : tf.__version__,
                                            'CLASS_NAMES'             : CLASS_NAMES
                                                  
                                            }                     
                                } )
    
    # @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
    
    YOLO_SPEC     = NEW_CONFIG['YOLO_SPEC']
    
    TRAIN_CONFIG  = NEW_CONFIG['TRAIN_CONFIG']
    
    BATCHING_SPEC = TRAIN_CONFIG['BATCHING']
    LOGGING       = TRAIN_CONFIG['LOGGING']
    
    # @@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@

    if IMG_DATASET is not None:
        
        if TASK == DTT: 
            
            pipe_builder = YoloPipe( user_defined_aug_pipe, BATCHING_SPEC, NUM_CLASSES, YOLO_SPEC )
        
        else: # I.E. CLS 
            
            pipe_builder = ImgPipe( user_defined_aug_pipe, BATCHING_SPEC, NUM_CLASSES )
        
        try:
            
            print('Attempting to load dataset in memory ...')
            
            train_samples, train_labels, eval_samples, eval_labels = image_dataset.load_in_memory()
            
            print('Succesfully loaded dataset into memory !')
            
            def train_input_pipe():
                
                return pipe_builder.build_mem_pipe(train_samples, train_labels, train_flag = True)
            
            def eval_input_pipe():
                
                return pipe_builder.build_mem_pipe(eval_samples, eval_labels, train_flag = False)
            
            print('Finished building input pipes from for both training and evaluation dataset found in memory')
            
        except MemThresholdReached as mtr:
            
            print( mtr )
            
            if TASK == DTT: 
                
                converter = YoloConverter( image_dataset )
                
            else: # I.E CLS
                
                converter = ImgConverter( image_dataset )
            
            path_to_tfr_dataset = converter.convert()
        
            def train_input_pipe():
                
                return pipe_builder.build_tfr_pipe( join( path_to_tfr_dataset, TRAIN_TFR_NAME ), INPUT_SHAPE, train_flag = True )
            
            def eval_input_pipe():
            
                return pipe_builder.build_tfr_pipe( join( path_to_tfr_dataset, EVAL_TFR_NAME ), INPUT_SHAPE, train_flag = False )
    
            print('Finished building input pipes for both training and evaluation dataset that was just serialized to disk')
    
    else:
        
        if TASK == DTT:
            
            pipe_builder = YoloPipe( user_defined_aug_pipe, BATCHING_SPEC, NUM_CLASSES, YOLO_SPEC )
        
        else: # I.E. CLS
            
            pipe_builder = ImgPipe( user_defined_aug_pipe, BATCHING_SPEC, NUM_CLASSES )
        
        def train_input_pipe():
            
            return pipe_builder.build_tfr_pipe( join( TFR_DATASET, TRAIN_TFR_NAME ), INPUT_SHAPE, train_flag = True )
        
        def eval_input_pipe():
    
            return pipe_builder.build_tfr_pipe( join( TFR_DATASET, EVAL_TFR_NAME ), INPUT_SHAPE, train_flag = False )

        print('Finished building input pipes for both training and evaluation dataset found in tfr serialized form on disk')

    TRAIN_STEPS = int( NUM_TRAIN_EXAMPLES / BATCHING_SPEC['BATCH_SIZE'] ) * TRAIN_CONFIG['LOOP_OVER_TRAIN_SET_COUNT']

# &&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&&  


def main():

    logging.getLogger().setLevel( logging.INFO )

    session_config = None

    _os = platform()
    if _os[0] == 'W' : # for Windows

        session_config = tf.ConfigProto()
        session_config.gpu_options.allow_growth = True
   
    RUN_CONFIG = tf.estimator.RunConfig( save_summary_steps     = TRAIN_STEPS, # this way, we only end up with metrics graphed in sync, I.E for each value of metric in train step graphed, there will be one value of that metric in eval step
                                         log_step_count_steps   = 1, 
                                         keep_checkpoint_max    = LOGGING[ 'KEEP_CHECKPOINT_MAX' ],
                                         save_checkpoints_steps = LOGGING[ 'SAVE_CHECKPOINTS_STEPS' ],
                                         session_config         = session_config )
    
    params = { 
              
                'YOLO_SPEC'     : YOLO_SPEC, 
                'MODEL_FUNCTION': user_defined_model_function,
                'NUM_CLASSES'   : NUM_CLASSES,
                'TASK'          : TASK
              
        }
    
    if TASK == DTT:
        
        estimator = YoloEstimator( 
        
                                    model_fn = build_estimator_model_function, 
                                    params = params, 
                                    model_dir = DUMP_DIR, 
                                    config = RUN_CONFIG 
                                        
                            )
        
    else: # I.E. CLS
        
        estimator = ClassifierEstimator( 
        
                                            model_fn = build_estimator_model_function, 
                                            params = params, 
                                            model_dir = DUMP_DIR, 
                                            config = RUN_CONFIG 
                                        
                                )
    
    if MODE == TRAIN:
            
        while True:
            
            estimator.train( input_fn = train_input_pipe, steps = TRAIN_STEPS ) # goes through all train set, doing weight update for every batch, to do an evaluation at the end and REPEAT
            out_dict = estimator.evaluate( input_fn = eval_input_pipe )
    
            '''
            
                Note: 
                
                Estimator.evaluate() does spit out other metrics also, the have to be passed into the eval_metric_ops arg for the EstimatorSpec object.
                Seems like the loss op is added implicitly by the Estimator API
            
                Find a way to detect platteaus so that we can do cross validation
            
            '''

    else: # I.E PREDICT since EXPORT doesn't even reach this far
        
        if TASK == CLS:

            estimator.predict_wrapper(SRC_DIR, DST_DIR, EXTENSION, CONFIG['TRAIN_INFO'])

        else:

            estimator.predict_wrapper(SRC_DIR, DST_DIR, EXTENSION, CONFIG['TRAIN_INFO'], YOLO_SPEC)
