import tensorflow as tf
import numpy      as np

from .custom_exceptions import MissingModelFile, MissingConfigFile, MissingTrainTFR, MissingEvalTFR, ModelDatasetMissmatch

from .project_globals import TRAIN, PREDICT, EXPORT, CLS, DTT
from .project_globals import TRAIN_TFR_NAME, EVAL_TFR_NAME, METADATA_FILE_NAME, CONFIG_FILE_NAME, MODEL_FILE_NAME

from .project_globals import MODE_ARG, TASK_ARG, CONFIG_ARG, DUMP_ARG, PLATTEAU_TOLERANCE_ARG, IMG_DATASET_ARG, COLOR_ARG, EXTENSION_ARG
from .project_globals import PERCENT_ARG, THRESHOLD_ARG, TFR_DATASET_ARG, CHECKPOINT_DIR_ARG, EXPORT_MODEL_DIR_ARG, EXPORT_MODEL_FILE_NAME_ARG
from .project_globals import SRC_DIR_ARG, DST_DIR_ARG
from .project_globals import TRAIN_DS_PERCENT_MIN, TRAIN_DS_PERCENT_MAX, MEM_THRESHOLD_PERCENT_MIN, MEM_THRESHOLD_PERCENT_MAX
from .project_globals import INPUT_OP_NAME, OUTPUT_OP_NAME

from json import load, dump
from os import listdir
from os.path import isdir, isfile, join

from tensorflow.python.tools.optimize_for_inference_lib import optimize_for_inference


def export_inference_model(ckpt_dir, model_dir, model_name):

    latest_checkpoint_file = tf.train.latest_checkpoint(ckpt_dir)

    latest_meta_file = latest_checkpoint_file + '.meta'

    graph = tf.Graph()

    with tf.Session(graph = graph) as sess:
        
        saver = tf.train.import_meta_graph( latest_meta_file )

        saver.restore( sess, latest_checkpoint_file )

        frozen_graph_def = tf.graph_util.convert_variables_to_constants(
            sess, 
            graph.as_graph_def(), 
            [ OUTPUT_OP_NAME ]
        ) 

        inference_graph_def = optimize_for_inference(

            frozen_graph_def, 
            input_node_names = [INPUT_OP_NAME ], 
            output_node_names = [OUTPUT_OP_NAME ], 
            placeholder_type_enum = tf.float32.as_datatype_enum,
            toco_compatible = False # look into what this param is about
    
        )

        inference_model_file_name = join( model_dir, model_name )

        with tf.gfile.GFile( inference_model_file_name, "wb" ) as f:
            f.write( inference_graph_def.SerializeToString() )

    print('Inference graph can be found at ', inference_model_file_name)
    exit(0)

def interval_overlap(interval_a, interval_b):
    
    x1, x2 = interval_a
    x3, x4 = interval_b

    if x3 < x1:
        if x4 < x1:
            return 0
        else:
            return min(x2,x4) - x1
    else:
        if x2 < x3:
             return 0
        else:
            return min(x2,x4) - x3

def bbox_iou(box1, box2):
    
    # xmin, ymin, xmax, ymax is the original implementation using BoundBox object
    
    intersect_w = interval_overlap( [ box1[0], box1[2] ], [ box2[0], box2[2] ] )
    intersect_h = interval_overlap( [ box1[1], box1[3] ],  [ box2[1], box2[3] ] )  
    
    intersect = intersect_w * intersect_h

    w1, h1 = box1[2] - box1[0], box1[3] - box1[1]
    w2, h2 = box2[2] - box2[0], box2[3] - box2[1]
    
    union = w1 * h1 + w2 * h2 - intersect
    
    return float(intersect) / union

def check_args_validity(args):
    
    print('Checking argument validity ...')
    
    # doing this because i'd rather unpack everything when returning from this call, main stays cleaner
    
    MODE = args[MODE_ARG]
    TASK = args[TASK_ARG]
    CONFIG_DIR = args[CONFIG_ARG]
    DUMP_DIR = args[DUMP_ARG]
    PLATTEAU_TOLERANCE = args[PLATTEAU_TOLERANCE_ARG]
    IMG_DATASET = args[IMG_DATASET_ARG]
    COLOR = args[COLOR_ARG]
    EXTENSION = args[EXTENSION_ARG]
    PERCENT = args[PERCENT_ARG]
    THRESHOLD = args[THRESHOLD_ARG]
    TFR_DATASET = args[TFR_DATASET_ARG]
    SRC_DIR = args[SRC_DIR_ARG]
    DST_DIR = args[DST_DIR_ARG]
    CHECKPOINT_DIR = args[CHECKPOINT_DIR_ARG]
    EXPORT_MODEL_DIR = args[EXPORT_MODEL_DIR_ARG]
    EXPORT_MODEL_FILE_NAME = args[EXPORT_MODEL_FILE_NAME_ARG]
    
    user_arg_combo = set()
    
    for key, value in args.items():
        
        if key == MODE_ARG: continue
        
        if value is not None: user_arg_combo.add(key)
    
    TRAIN_COMBO = { TASK_ARG, CONFIG_ARG, DUMP_ARG }
        
    IMG_DATASET_COMBO = { IMG_DATASET_ARG, COLOR_ARG, EXTENSION_ARG, THRESHOLD_ARG, PERCENT_ARG }
    TFR_DATASET_COMBO = { TFR_DATASET_ARG }
    
    
    train_img_dataset_flag = user_arg_combo == TRAIN_COMBO.union( IMG_DATASET_COMBO )
    train_tfr_dataset_flag = user_arg_combo == TRAIN_COMBO.union( TFR_DATASET_COMBO )
    
    if MODE == TRAIN:

        if train_img_dataset_flag and train_tfr_dataset_flag : # both combos are valid --> 1 and 1 = 1
            
            raise ValueError( 'You can have either ' + str( TRAIN_COMBO + IMG_DATASET_COMBO ) + ' for training with an image dataset or\n' +
                        
                                str( TRAIN_COMBO + TFR_DATASET_COMBO ) + ' for training with a tfr dataset.'
                              
                        )
    
        if not( train_img_dataset_flag or train_tfr_dataset_flag ): # both combos are invalid --> 0 nor 0 = 1
            
            err_msg = (
            
                'Invalid optional argument combo! ' + 
                '\nFor training with an image dataset, the following combo si required: ' + str( TRAIN_COMBO.union( IMG_DATASET_COMBO ) ) +
                '\nFor training with a tfr dataset, the following combo is required: ' + str( TRAIN_COMBO.union( TFR_DATASET_COMBO ) )

            )
            
            raise ValueError( err_msg )

    if MODE == PREDICT:
          
        PREDICT_COMBO = { TASK_ARG, CONFIG_ARG, DUMP_ARG, SRC_DIR_ARG, DST_DIR_ARG, EXTENSION_ARG }
        
        if user_arg_combo != PREDICT_COMBO:
            
            err_msg = 'You can only use the following combo with ' + PREDICT + ' mode: ' + str(PREDICT_COMBO) 
            
            raise ValueError(err_msg)
        
        if not isdir(SRC_DIR):
            
            err_msg = 'Predict source directory does not exist'
            
            raise ValueError(err_msg)
            
        if not isdir(DST_DIR):
            
            err_msg = 'Predict destination directory does not exist'
            
            raise ValueError(err_msg)
        
        
    if MODE == EXPORT:
        
        EXPORT_COMBO = { CHECKPOINT_DIR_ARG, EXPORT_MODEL_DIR_ARG, EXPORT_MODEL_FILE_NAME_ARG }
        
        if user_arg_combo != EXPORT_COMBO:
            
            raise ValueError('You can only use the followig combo with ' + EXPORT + ' mode: ' + str(EXPORT_COMBO) )
    
    
    if MODE in [ TRAIN, PREDICT ]:
        
        # for EXPORT mode we don't specify config_dir or dump_dir, call to isdir will fail with None type arg
        
        if not isdir(CONFIG_DIR) : raise IOError( 'Config directory does not exist !' )
        if not isdir(DUMP_DIR) : raise IOError('Dump directory does not exist !')
        
        config_files = listdir( CONFIG_DIR )
        
        if MODEL_FILE_NAME not in config_files: raise MissingModelFile('Expected ' + MODEL_FILE_NAME + ' in ' + CONFIG_DIR)
        if CONFIG_FILE_NAME not in config_files: raise MissingConfigFile('Expected ' + CONFIG_FILE_NAME + ' in ' + CONFIG_DIR)
    
    if train_img_dataset_flag:
        
        '''
        
            Note: 
              
                To avoid type error when casting COLOR variable from None type to bool in some scenarios when returning out of func, 
                make it now a bool, when we know it's a valid arg, else it will be returned as None with no prob.
        
                Maybe there is a way to pass an explicit bool value in argparse, atm you pass int 0 or 1 to be later cast to bool
                i don't wanna do branching based on the string version of it
        
        '''
        
        if not isdir( IMG_DATASET ): 
            
            raise IOError('Image dataset directory does not exist !')
        
        if COLOR is None:
            
            raise ValueError( '--' + COLOR_ARG + ' is required' ) 
        
        COLOR = bool( COLOR )
        
        if EXTENSION is None:
            
            raise ValueError( '--' + EXTENSION_ARG + ' is reuired')
        
        if PERCENT is None :
            
            raise ValueError( '--' + PERCENT_ARG + ' is required')
        
        if not ( TRAIN_DS_PERCENT_MIN <= PERCENT <= TRAIN_DS_PERCENT_MAX ):
            
            raise ValueError( '--' + PERCENT_ARG + ' value should be between ' + str( TRAIN_DS_PERCENT_MIN ) + ' and ' + str( TRAIN_DS_PERCENT_MAX ) ) 
        
        if THRESHOLD is None:
            
            raise ValueError( '--' + THRESHOLD_ARG + ' is required' )
    
        if not( MEM_THRESHOLD_PERCENT_MIN <= THRESHOLD <= MEM_THRESHOLD_PERCENT_MAX ):
            
            raise ValueError( '--' + THRESHOLD_ARG + ' value should be between ' + str( MEM_THRESHOLD_PERCENT_MIN ) + ' and ' + str( MEM_THRESHOLD_PERCENT_MAX ) )
    
    if train_tfr_dataset_flag:
        
        if not isdir( TFR_DATASET ): 
            
            raise IOError('TFR dataset directory does not exist !')
        
        tfr_dataset_dir_contents = listdir( TFR_DATASET )
    
        with open( join( TFR_DATASET, METADATA_FILE_NAME ), 'r' ) as handler:
        
            metadata = load(handler)
    
        darknet_ds_flag = metadata['DARKNET_DS_FLAG']
    
        if TASK == CLS and darknet_ds_flag: raise ModelDatasetMissmatch('Cannot train a classifier on a darknet dataset ment for a yolo detector')
        if TASK == DTT and not darknet_ds_flag : raise modelDatasetMissmatch('Cannot train a yolo detector on a regular image dataset ment for a classifier')
    
        if TRAIN_TFR_NAME not in tfr_dataset_dir_contents     : raise MissingTrainTFR('Expected name should be ' + TRAIN_TFR_NAME)
        if EVAL_TFR_NAME not in tfr_dataset_dir_contents      : raise MissingEvalTFR('Expected name should be ' + EVAL_TFR_NAME)
        if METADATA_FILE_NAME not in tfr_dataset_dir_contents : raise MissingMetaDataFile('Expected name should be ' + METADATA_FILE_NAME)
    
    return MODE, TASK, CONFIG_DIR, DUMP_DIR, PLATTEAU_TOLERANCE, IMG_DATASET, COLOR, EXTENSION, PERCENT, THRESHOLD, TFR_DATASET, SRC_DIR, DST_DIR, CHECKPOINT_DIR, EXPORT_MODEL_DIR, EXPORT_MODEL_FILE_NAME

def validate_darknet_dataset(path_to_ds_root_dir, img_extension):
    
    pair_missmatch_dir = join(path_to_ds_root_dir, 'MISMATCHES')
    
    if not isdir(pair_missmatch_dir): mkdir (pair_missmatch_dir)

    
    all_images = glob( join( path_to_ds_root_dir, '*' + img_extension ) )
    all_annotations = glob( join( path_to_ds_root_dir, '*' + '.txt' ) )
    
    # TODO: relocate either images with no coresponding txt file or viceversa to the MISMATCHES directory
            
def update_config(config_file, entry, key = None):
    
    '''
    
        Update or append specific entries in config.json
    
    '''
    
    with open( config_file, 'r' ) as config_json:
        
        config_data = load( config_json )

    if key is None: config_data.update( entry )
    else: config_data[ key ].update( entry ) 

    with open( config_file, 'w') as config_json:
    
        dump(config_data, config_json, indent = 5)
        
    return config_data 


def yolo_loss(y_pred, y_true, true_boxes, yolo_spec):

    #tmp = tf.Print(y_true, [tf.shape(y_true), tf.shape(true_boxes), tf.shape(y_pred)], summarize = 50)

    num_anchors = int( len( yolo_spec['ANCHORS'] )  / 2 )

    mask_shape = tf.shape( y_true )[:4]
    batch_size = tf.shape(y_pred)[0] # since it can differ, better compute it at runtime, for instance if num_examples % batch_size != 0 we are screwed, might have to make batching to be based on a NUM_BATCHES system relative to the dataset

    cell_x = tf.to_float( tf.reshape( tf.tile( tf.range( yolo_spec['GRID_W'] ), [ yolo_spec['GRID_H'] ] ), ( 1, yolo_spec['GRID_H'], yolo_spec['GRID_W'], 1, 1 ) ) )
    
    cell_y = tf.to_float( tf.reshape( tf.tile( tf.reshape( tf.range( yolo_spec['GRID_H'] ), ( yolo_spec['GRID_H'], 1 ) ), [ 1, yolo_spec['GRID_W'] ] ), ( 1, yolo_spec['GRID_H'], yolo_spec['GRID_W'], 1, 1 ) ) ) 

    cell_grid = tf.tile( tf.concat( [ cell_x, cell_y ], -1 ), [ batch_size, 1, 1, num_anchors, 1 ] )

    coord_mask = tf.zeros( mask_shape )
    conf_mask = tf.zeros( mask_shape )
    class_mask = tf.zeros( mask_shape )

    seen = tf.Variable(0.)
    total_recall = tf.Variable(0.)

    """
    Adjust prediction
    """
    ### adjust x and y
    pred_box_xy = tf.sigmoid(y_pred[..., :2]) + cell_grid

    ### adjust w and h
    pred_box_wh = tf.exp(y_pred[..., 2:4]) * np.reshape( yolo_spec['ANCHORS'], [1, 1, 1, num_anchors, 2] )

    ### adjust confidence
    pred_box_conf = tf.sigmoid(y_pred[..., 4])

    ### adjust class probabilities
    pred_box_class = y_pred[..., 5:]

    """
    Adjust ground truth
    """
    ### adjust x and y
    true_box_xy = y_true[..., 0:2]  # relative position to the containing cell

    ### adjust w and h
    true_box_wh = y_true[..., 2:4]  # number of cells accross, horizontally and vertically

    ### adjust confidence
    true_wh_half = true_box_wh / 2.
    true_mins = true_box_xy - true_wh_half
    true_maxes = true_box_xy + true_wh_half

    pred_wh_half = pred_box_wh / 2.
    pred_mins = pred_box_xy - pred_wh_half
    pred_maxes = pred_box_xy + pred_wh_half

    intersect_mins = tf.maximum(pred_mins, true_mins)
    intersect_maxes = tf.minimum(pred_maxes, true_maxes)
    intersect_wh = tf.maximum(intersect_maxes - intersect_mins, 0.)
    intersect_areas = intersect_wh[..., 0] * intersect_wh[..., 1]

    true_areas = true_box_wh[..., 0] * true_box_wh[..., 1]
    pred_areas = pred_box_wh[..., 0] * pred_box_wh[..., 1]

    union_areas = pred_areas + true_areas - intersect_areas
    iou_scores = tf.truediv(intersect_areas, union_areas)

    true_box_conf = iou_scores * y_true[..., 4]

    ### adjust class probabilities
    true_box_class = tf.argmax(y_true[..., 5:], -1)

    """
    Determine the masks
    """
    ### coordinate mask: simply the position of the ground truth boxes (the predictors)
    coord_mask = tf.expand_dims(y_true[..., 4], axis=-1) * yolo_spec['COORD_SCALE']

    ### confidence mask: penelize predictors + penalize boxes with low IOU
    # penalize the confidence of the boxes, which have IOU with some ground truth box < 0.6
    true_xy = true_boxes[..., 0:2]
    true_wh = true_boxes[..., 2:4]

    true_wh_half = true_wh / 2.
    true_mins = true_xy - true_wh_half
    true_maxes = true_xy + true_wh_half

    pred_xy = tf.expand_dims(pred_box_xy, 4)
    pred_wh = tf.expand_dims(pred_box_wh, 4)

    pred_wh_half = pred_wh / 2.
    pred_mins = pred_xy - pred_wh_half
    pred_maxes = pred_xy + pred_wh_half

    intersect_mins = tf.maximum(pred_mins, true_mins)
    intersect_maxes = tf.minimum(pred_maxes, true_maxes)
    intersect_wh = tf.maximum(intersect_maxes - intersect_mins, 0.)
    intersect_areas = intersect_wh[..., 0] * intersect_wh[..., 1]

    true_areas = true_wh[..., 0] * true_wh[..., 1]
    pred_areas = pred_wh[..., 0] * pred_wh[..., 1]

    union_areas = pred_areas + true_areas - intersect_areas
    iou_scores = tf.truediv(intersect_areas, union_areas)

    best_ious = tf.reduce_max(iou_scores, axis=4)
    conf_mask = conf_mask + tf.to_float(best_ious < 0.6) * (1 - y_true[..., 4]) * yolo_spec['NO_OBJECT_SCALE']

    # penalize the confidence of the boxes, which are reponsible for corresponding ground truth box
    conf_mask = conf_mask + y_true[..., 4] * yolo_spec['OBJECT_SCALE']

    ### class mask: simply the position of the ground truth boxes (the predictors)
    class_mask = y_true[..., 4] * tf.gather( yolo_spec['CLASS_WEIGHTS'], true_box_class ) * yolo_spec['CLASS_SCALE']

    """
    Warm-up training
    """
    no_boxes_mask = tf.to_float(coord_mask < yolo_spec['COORD_SCALE'] / 2.)
    seen = tf.assign_add(seen, 1.)

    true_box_xy, true_box_wh, coord_mask = tf.cond(tf.less(seen, yolo_spec['WARM_UP_BATCHES']),
                                                   lambda: [true_box_xy + (0.5 + cell_grid) * no_boxes_mask,
                                                            true_box_wh + tf.ones_like(true_box_wh) * np.reshape(
                                                                yolo_spec['ANCHORS'], [1, 1, 1, num_anchors, 2]) * no_boxes_mask,
                                                            tf.ones_like(coord_mask)],
                                                   lambda: [true_box_xy,
                                                            true_box_wh,
                                                            coord_mask])

    """
    Finalize the loss
    """
    nb_coord_box = tf.reduce_sum(tf.to_float(coord_mask > 0.0))
    nb_conf_box = tf.reduce_sum(tf.to_float(conf_mask > 0.0))
    nb_class_box = tf.reduce_sum(tf.to_float(class_mask > 0.0))

    loss_xy = tf.reduce_sum(tf.square(true_box_xy - pred_box_xy) * coord_mask) / (nb_coord_box + 1e-6) / 2.
    loss_wh = tf.reduce_sum(tf.square(true_box_wh - pred_box_wh) * coord_mask) / (nb_coord_box + 1e-6) / 2.
    loss_conf = tf.reduce_sum(tf.square(true_box_conf - pred_box_conf) * conf_mask) / (nb_conf_box + 1e-6) / 2.
    loss_class = tf.nn.sparse_softmax_cross_entropy_with_logits(labels=true_box_class, logits=pred_box_class)
    loss_class = tf.reduce_sum(loss_class * class_mask) / (nb_class_box + 1e-6)

    # the add op version of the loss yields ' Failed to run optimizer ArithmeticOptimizer '
    #loss = loss_xy + loss_wh + loss_conf + loss_class
    
    # this one doesn't but i can't vouch for wether it's working as intended or not yet !!! Can't see why it would not tho :D
    loss = tf.reduce_sum( [ loss_xy, loss_wh, loss_conf, loss_class ] )

    nb_true_box = tf.reduce_sum(y_true[..., 4])
    nb_pred_box = tf.reduce_sum(tf.to_float(true_box_conf > 0.5) * tf.to_float(pred_box_conf > 0.3))

    """
    Debugging code
    """
    
    # current_recall = nb_pred_box / (nb_true_box + 1e-6)
    # total_recall = tf.assign_add(total_recall, current_recall)
    #
    # loss = tf.Print(loss, [tf.zeros((1))], message='Dummy Line \t', summarize=1000)
    # loss = tf.Print(loss, [loss_xy], message='Loss XY \t', summarize=1000)
    # loss = tf.Print(loss, [loss_wh], message='Loss WH \t', summarize=1000)
    # loss = tf.Print(loss, [loss_conf], message='Loss Conf \t', summarize=1000)
    # loss = tf.Print(loss, [loss_class], message='Loss Class \t', summarize=1000)
    # loss = tf.Print(loss, [loss], message='Total Loss \t', summarize=1000)
    # loss = tf.Print(loss, [current_recall], message='Current Recall \t', summarize=1000)
    # loss = tf.Print(loss, [total_recall / seen], message='Average Recall \t', summarize=1000)

    return loss


def append_logits_layer(current_output, num_classes):
    
    return tf.layers.dense(inputs = current_output, units = num_classes, activation = None, name = 'logits')

def append_yolo_layer(current_output, num_anchors, num_classes, grid_w, grid_h):

    # might have different implicit values for the parameters compared to the keras implementation of Conv2D, have to do some checking !!!
    
    yolo_layer_output = tf.layers.conv2d( inputs = current_output, 
                                          filters = num_anchors * (4 + 1 + num_classes), 
                                          kernel_size = 1, 
                                          padding = 'same', 
                                          strides = 1, 
                                          activation = None, 
                                          name = 'yolo_layer')
    
    # freggin think reshaped the entire batch into [grid_w, grid_h, num_anchors, 4 + 1 + num_classes ] if it werent for the extra dim -1
    return tf.reshape( yolo_layer_output, [ -1, grid_h, grid_w, num_anchors, 4 + 1 + num_classes ], name = OUTPUT_OP_NAME )


def get_max_pool_count(graph):
    
    all_ops = graph.get_operations()
    
    all_max_pools = [ op for op in all_ops if 'MaxPool' in op.name ]
    
    return len( all_max_pools )

def build_estimator_model_function(features, labels, mode, params):

    '''
    
        Some refactoring might be needed, have to check what can be done 
    
    '''

    YOLO_MODEL = params['TASK'] == DTT
    
    user_defined_model_functon = params['MODEL_FUNCTION']
    
    model_input = tf.identity(features, name = INPUT_OP_NAME )
    
    current_model_output, user_defined_optimizer_op = user_defined_model_functon(model_input, train_flag = ( mode == tf.estimator.ModeKeys.TRAIN ) )
    
    if YOLO_MODEL:
 
        prediction = append_yolo_layer( current_model_output, 
                                        int ( len(  params['YOLO_SPEC']['ANCHORS']  ) / 2 ) ,
                                        params['NUM_CLASSES'],
                                        params['YOLO_SPEC']['GRID_W'],
                                        params['YOLO_SPEC']['GRID_H']
                                    )
    
    else: # I.E. CLS
        
        logits = append_logits_layer( current_model_output, params['NUM_CLASSES'] )
        
        prediction = tf.nn.softmax(logits = logits, name = OUTPUT_OP_NAME)
    
    if mode == tf.estimator.ModeKeys.PREDICT:
        
        spec = tf.estimator.EstimatorSpec( mode = mode, predictions = prediction )
        
    else :
        
        if YOLO_MODEL:
            
            # TODO: add IOU and other detector specific metric logging capabilities
            metrics = None
            
            '''
            
            labels[0] -> y_batch
            labels[1] -> b_batch
            
            '''
            
            
            loss = yolo_loss(prediction, labels[0], labels[1], params['YOLO_SPEC'] )
            
            
        else:
        
            # tf.nn.softmax_cross_entropy_with_logits_v2 seems to be replacing the depricated version which is currently beeing used, have to look it up first
            loss = tf.nn.softmax_cross_entropy_with_logits( labels = labels, logits = logits)
            
            ## 'accuracy' --> [UPDATE_OP, VALUE]
            accuracy = tf.metrics.accuracy( labels = tf.argmax( labels, 1 ), predictions = tf.argmax( prediction, 1 ), name = 'acc_metric')
            metrics = { 'accuracy' : accuracy }
            
            # maybe add histograms for weight matrices, to get more insight on wether overfitting is a problem. or to debug regularization techniques.
            # see if they were applied properly or had any effect even so.
            
            '''
            
                use a custom summary writer for train loss monioring with increased frequency, lets say
                every weight update, as opposed to the logging done after each train / eval set pass
            
                same scalar name, just saved in different dir, that will add it in the same graph but with a different color :D
            
                this applied for any metric, might have to do some experimenting with this
            
            '''
            
            tf.summary.scalar( name = 'accuracy', tensor = accuracy[1] )
            tf.summary.merge_all()
        
        loss = tf.reduce_mean( loss )
        
        optimizer_op = user_defined_optimizer_op.minimize( loss = loss, global_step = tf.train.get_global_step() )
        
        spec = tf.estimator.EstimatorSpec( mode = mode, loss = loss, train_op = optimizer_op , eval_metric_ops = metrics )
        
    return spec

def get_kernel_size(factor):
	#Find the kernel size given the desired factor of upsampling 
	return 2 * factor + factor % 2

def upsample_filt(size):
	#Make a 2D bilinear kernel suitable for upsampling of the given (h, w) size;

	factor = (size+1) // 2
	if size % 2 == 1:
		center = factor - 1
	else:
		center = factor - 0.5
	return (1 - abs(og[0] - center) / factor) * (1- abs(og[1] - center) / factor)

def biliniar_upsample_weigts(factor, number_of_classes):
	#create weights matrix for transpose convolution with biliniar filter initializations

	filter_size = get_kernel_size(factor)
	weights = np.zeros((filter_size, filter_size, number_of_classes, number_of_classes), dtype = np.float32)
	upsample_kernel = upsample_filt(filter_size)

	for i in range(number_of_classes):
		weights[:, :, i, i] = upsample_kernel
	return weights







